# -*- coding: iso-8859-1 -*-
"""
    MoinMoin - PostMessage Action

    @copyright: 2012, 2013 by Paul Boddie <paul@boddie.org.uk>
    @license: GNU GPL (v2 or later), see COPYING.txt for details.
"""

from MoinMoin.Page import Page
from MoinMoin.PageEditor import PageEditor
from MoinMoin.log import getLogger
from MoinMoin.user import User
from MoinSupport import *
from MoinMessage import GPG, Message, MessageStore, MoinMessageError
from email.parser import Parser

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

Dependencies = ['pages']

class PostMessage:

    "A posted message handler."

    def __init__(self, pagename, request):

        """
        On the page with the given 'pagename', use the given 'request' when
        reading posted messages, modifying the Wiki.
        """

        self.pagename = pagename
        self.request = request
        self.page = Page(request, pagename)

    def do_action(self):
        request = self.request
        content_length = getHeader(request, "Content-Length", "HTTP")
        if content_length:
            content_length = int(content_length)

        self.handle_message_text(request.read(content_length))

    def handle_message_text(self, message_text):

        "Handle the given 'message_text'."

        message = Parser().parse(StringIO(message_text))
        self.handle_message(message)

    def handle_message(self, message):

        "Handle the given 'message'."

        request = self.request
        mimetype = message.get_content_type()
        encoding = message.get_content_charset()

        # Detect PGP/GPG-encoded payloads.
        # See: http://tools.ietf.org/html/rfc3156

        if mimetype == "multipart/signed" and \
            message.get_param("protocol") == "application/pgp-signature":

            self.handle_signed_message(message)

        elif mimetype == "multipart/encrypted" and \
            message.get_param("protocol") == "application/pgp-encrypted":

            self.handle_encrypted_message(message)

        # Reject unsigned payloads.

        else:
            writeHeaders(request, "text/plain", getMetadata(self.page), "415 Unsupported Media Type")
            request.write("Only PGP/GPG-signed payloads are supported.")

    def handle_encrypted_message(self, message):

        "Handle the given encrypted 'message'."

        request = self.request

        try:
            declaration, content = message.get_payload()
        except ValueError:
            writeHeaders(request, "text/plain", getMetadata(self.page), "415 Unsupported Media Type")
            request.write("There must be a declaration and a content part for encrypted uploads.")
            return

        # Verify the message format.

        if content.get_content_type() != "application/octet-stream":
            writeHeaders(request, "text/plain", getMetadata(self.page), "415 Unsupported Media Type")
            request.write("Encrypted data must be provided as application/octet-stream.")
            return

        homedir = self.get_homedir()
        if not homedir:
            return

        gpg = GPG(homedir)

        # Get the decrypted message text.

        try:
            text = gpg.decryptMessage(content.get_payload())

            # Log non-fatal errors.

            if gpg.errors:
                getLogger(__name__).warning(gpg.errors)

            # Handle the embedded message.

            self.handle_message_text(text)

        # Otherwise, reject the unverified message.

        except MoinMessageError:
            writeHeaders(request, "text/plain", getMetadata(self.page), "403 Forbidden")
            request.write("The message could not be decrypted.")

    def handle_signed_message(self, message):

        "Handle the given signed 'message'."

        request = self.request

        # NOTE: RFC 3156 states that signed messages should employ a detached
        # NOTE: signature but then shows "BEGIN PGP MESSAGE" for signatures
        # NOTE: instead of "BEGIN PGP SIGNATURE".
        # NOTE: The "micalg" parameter is currently not supported.

        try:
            content, signature = message.get_payload()
        except ValueError:
            writeHeaders(request, "text/plain", getMetadata(self.page), "415 Unsupported Media Type")
            request.write("There must be a content part and a signature for signed uploads.")
            return

        # Verify the message format.

        if signature.get_content_type() != "application/pgp-signature":
            writeHeaders(request, "text/plain", getMetadata(self.page), "415 Unsupported Media Type")
            request.write("Signature data must be provided in the second part as application/pgp-signature.")
            return

        homedir = self.get_homedir()
        if not homedir:
            return

        gpg = GPG(homedir)

        # Verify the message.

        try:
            fingerprint, identity = gpg.verifyMessage(signature.get_payload(), content.as_string())

            # Map the fingerprint to a Wiki user.

            old_user = None
            request = self.request

            try:
                if fingerprint:
                    gpg_users = getWikiDict(
                        getattr(request.cfg, "moinmessage_gpg_users_page", "MoinMessageUserDict"),
                        request
                        )

                    # With a user mapping and a fingerprint corresponding to a known
                    # user, temporarily switch user in order to make the edit.

                    if gpg_users and gpg_users.has_key(fingerprint):
                        old_user = request.user
                        request.user = User(request, auth_method="gpg", auth_username=gpg_users[fingerprint])

                # Log non-fatal errors.

                if gpg.errors:
                    getLogger(__name__).warning(gpg.errors)

                # Handle the embedded message.

                self.handle_message_content(content)

            # Restore any user identity.

            finally:
                if old_user:
                    request.user = old_user

        # Otherwise, reject the unverified message.

        except MoinMessageError:
            writeHeaders(request, "text/plain", getMetadata(self.page), "403 Forbidden")
            request.write("The message could not be verified.")

    def handle_message_content(self, content):

        "Handle the given message 'content'."

        request = self.request

        # Interpret the content as one or more updates.

        message = Message()
        message.handle_message(content)

        for update in message.updates:

            # Handle a single part.

            if not update.is_multipart():
                self.handle_message_parts([update], update)

            # Or a collection of alternative representations for a single
            # update.

            else:
                self.handle_message_parts(update.get_payload(), update)

        # Default output.

        writeHeaders(request, "text/plain", getMetadata(self.page), "204 No Content")

    def handle_message_parts(self, parts, update):

        """
        Handle the given message 'parts', using the original 'update' to
        determine whether the content is to replace or update page content, or
        whether it will be placed in a message store.
        """

        # Handle the different update actions.
        # Update a message store for the page.

        if to_store(update):
            store = MessageStore(self.page)
            store.append(update)

        # Update the page.

        else:
            # NOTE: Should either choose preferred content types or somehow retain them
            # NOTE: all but present one at a time.

            body = []
            replace = to_replace(update)

            for part in parts:
                mimetype = part.get_content_type()
                encoding = part.get_content_charset()
                if mimetype == "text/moin":
                    body.append(part.get_payload())
                    if replace:
                        break

            if not replace:
                body.append(self.page.get_raw_body())

            page_editor = PageEditor(self.request, self.pagename)
            page_editor.saveText("\n\n".join(body), 0)

            # Refresh the page.

            self.page = Page(self.request, self.pagename)

    def get_homedir(self):

        "Locate the GPG home directory."

        homedir = getattr(self.request.cfg, "moinmessage_gpg_homedir")
        if not homedir:
            writeHeaders(request, "text/plain", getMetadata(self.page), "415 Unsupported Media Type")
            request.write("Encoded data cannot currently be understood. Please notify the site administrator.")
        return homedir

def to_replace(message):
    return message.get("Update-Action") == "replace"

def to_store(message):
    return message.get("Update-Action") == "store"

# Action function.

def execute(pagename, request):
    PostMessage(pagename, request).do_action() # instead of render

# vim: tabstop=4 expandtab shiftwidth=4
